unit MainUnit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, RbwZoomBox, contnrs, ExtCtrls, Buttons, RealListUnit,
  Menus;

type
  TfrmDEM2Image = class(TForm)
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    RbwZoomBox1: TRbwZoomBox;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Open1: TMenuItem;
    Refresh1: TMenuItem;
    Exit1: TMenuItem;
    Panel1: TPanel;
    btnCancel: TBitBtn;
    cbWhiteOcean: TCheckBox;
    sbPan: TSpeedButton;
    sbZoomExtents: TSpeedButton;
    sbZoomOut: TSpeedButton;
    sbZoomIn: TSpeedButton;
    sbZoom: TSpeedButton;
    Save1: TMenuItem;
    StatusBar1: TStatusBar;
    Help1: TMenuItem;
    Help2: TMenuItem;
    About1: TMenuItem;
    SaveasJPEG1: TMenuItem;
    cbRed: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RbwZoomBox1Paint(Sender: TObject);
    procedure sbZoomExtentsClick(Sender: TObject);
    procedure sbZoomInClick(Sender: TObject);
    procedure sbZoomOutClick(Sender: TObject);
    procedure RbwZoomBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure RbwZoomBox1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure RbwZoomBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure btnRefreshClick(Sender: TObject);
    procedure btnCancelClick(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure Refresh1Click(Sender: TObject);
    procedure sbZoomClick(Sender: TObject);
    procedure sbPanClick(Sender: TObject);
    procedure Save1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure Help2Click(Sender: TObject);
    procedure SaveasJPEG1Click(Sender: TObject);
    procedure FormResize(Sender: TObject);

  private
    DemFile : File;
 {$IFDEF WriteB}     DemOutput : TextFile; {$ENDIF}
    CharCount : integer;
    Columncount : integer;
    angle : double;
    ResolutionX, ResolutionY, ResolutionZ : double;
    Range : array[0..1] of double;
    Corners : TObjectList;
    Done : boolean;
    CoordInSec : boolean;
    CentralMeridianRadians : double;
    DEM_Map : TBitMap;
    PointSize : integer;
    RangeList : TRealList;
    CancelBitMap : boolean;
    function ReadCharacters(const Count: integer): string;
    procedure ReadRecordA(const GetCentralMeridian : boolean);
    procedure ReadRecordB;
    procedure ReadRestOfRecord;
    procedure ReadRecordC;
    function FortranStrToFloat(AString: string): double;
    function UTMCentralMeridianRadians(LatitudeSeconds,
      LongitudeSeconds: double): double;
    function Get_N(LatitudeRadians : double) : double;
    function Get_T(LatitudeRadians : double) : double;
    function Get_C(LatitudeRadians : double) : double;
    function Get_A(LongitudeRadians, LatitudeRadians,
      CentralMeridianRadians: double): double;
    function Get_M(LatitudeRadians: double): double;
    procedure GetCoordinates(const LatitudeRadians, LongitudeRadians,
      CentralMeridian: double; var X, Y: double);
    procedure DrawOnBitMap;
    function BM_Width: integer;
    function BM_Height: integer;
    procedure SetCornerCoordinates;
    procedure MakeBitMap(const ShouldZoomOut : boolean);
    procedure GetRange;
    { Private declarations }
  public
    { Public declarations }
  end;

  TElevationPoint = class(TRbwZoomPoint)
  private
    FElevation: double;
    FColor : TColor;
    procedure SetElevation(const Value: double);
    procedure Draw(const ABitMap: TBitmap; const PointSize: integer);
  published
    property Elevation : double read FElevation write SetElevation;
  end;

var
  frmDEM2Image: TfrmDEM2Image;

const  // Clarke ellipsoid
  LowerCase_a = 6378206.4; // in m
  e_square = 0.00676866;
  k_zero = 0.9996; // central scale factor
  e_prime_square = e_square/(1-e_square);

implementation

{$R *.DFM}

uses Math, CentralMeridianUnit, AboutUnit, JpegChoicesUnit;


Function FracToRainbow(Fraction : double) : TColor;
var
  Choice : integer;
begin
  Assert((Fraction>=0) and (Fraction<=1));
  fraction := Fraction*5;
  Choice := Trunc(fraction);
  fraction := Frac(fraction);
  case Choice of
    0:
      begin
        result := Round(Fraction*$FF)*$100 + $FF;       // R -> R+G
      end;
    1:
      begin
        result := $FF00 + Round((1-Fraction)*$FF);      // R+G -> G
      end;
    2:
      begin
        result := Round(Fraction*$FF)*$10000 + $FF00;  // G -> G+B
      end;
    3:
      begin
        result := $FF0000 + Round((1-Fraction)*$FF)*$100; // G+B -> B
      end;
    4:
      begin
        result := Round(Fraction*$FF) + $FF0000;          // B -> B+R
      end;
    5:
      begin
        result := $FF00FF;
      end;                                               // B+R
  else
    begin
      result := 0;
      Assert(False);
    end;
  end;
end;

procedure TfrmDEM2Image.ReadRestOfRecord;
var
  remainder : integer;
begin
  remainder := 1024 - CharCount mod 1024;
  if remainder <> 1024 then
  begin
    {$IFDEF WriteB} WriteLn(DemOutput,'remainder of record: ' + {$ENDIF} ReadCharacters(remainder){$IFDEF WriteB} ){$ENDIF};
  end;
  CharCount := 0;
end;

function TfrmDEM2Image.ReadCharacters(const Count : integer) : string;
var
  I : integer;
  AChar : Char;
begin
  SetLength(result,Count);
  CharCount := CharCount + Count;
  for I := 1 to Count do
  begin
    BlockRead(DemFile,AChar, 1);
    result[i] := AChar;
  end;
end;

procedure TfrmDEM2Image.SetCornerCoordinates;
var
  Index : integer;
  X, Y : double;
  CornerPoint : TRbwZoomPoint;
begin
  if CoordInSec then
  begin
    for Index := 0 to Corners.Count -1 do
    begin
      CornerPoint := Corners[Index] as TRbwZoomPoint;
      X := CornerPoint.X;
      Y := CornerPoint.Y;
      GetCoordinates(Y/60/60/180*PI, X/60/60/180*PI, CentralMeridianRadians, X, Y);
      CornerPoint.X := X;
      CornerPoint.Y := Y;
    end;
  end;
end;

procedure TfrmDEM2Image.GetRange;
begin
  RangeList.Sort;
  if RangeList.Count > 1 then
  begin
    Range[0] := RangeList[0];
    Range[1] := RangeList[RangeList.Count-1];
  end;
  RangeList.Clear;
end;

procedure TfrmDEM2Image.ReadRecordA(const GetCentralMeridian : boolean);
const
  CornerCount = 4;
var
  Index : integer;
  ColumncountString : string;
  AString : string;
//  temp : double;
  XString, YString : string;
  X, Y : double;
  CornerPoint : TRbwZoomPoint;
  MinOrMax : double;
begin
  CharCount := 0;
  {$IFDEF WriteB} WriteLn(DemOutput,'File name: ' + {$ENDIF} ReadCharacters(40) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Free Format Text: ' + {$ENDIF}  ReadCharacters(40) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Filler: ' +  {$ENDIF} ReadCharacters(55) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Process Code: ' + {$ENDIF}  ReadCharacters(1) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Filler: ' + {$ENDIF}  ReadCharacters(1) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Sectional Indicator: ' + {$ENDIF}  ReadCharacters(3) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'MC origin code: ' + {$ENDIF}  ReadCharacters(4) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'DEM level code: ' +  {$ENDIF} ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Code defining elevation pattern (regular or random): ' + {$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  AString := ReadCharacters(6);
  {$IFDEF WriteB} WriteLn(DemOutput,'Code defining ground planimetric reference system: ' + AString); {$ENDIF}
  CoordInSec := StrToInt(Trim(AString)) = 0;
  {$IFDEF WriteB} WriteLn(DemOutput,'Code defining zone in ground planimetric reference system: ' + {$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  for Index := 1 to 15 do
  begin
    {$IFDEF WriteB} WriteLn(DemOutput,'planimetric code '  + intToStr(Index) + ': ' + {$ENDIF}  ReadCharacters(24) {$IFDEF WriteB} ) {$ENDIF} ;
  end;
  {$IFDEF WriteB} WriteLn(DemOutput,'Code defining unit of measure for ground planimetric coordinates throughout the file: ' + {$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Code defining unit of measure for elevation coordinates throughout the file	: ' + {$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Number (n) of sides in the polygon which defines the coverage of the DEM file: ' + {$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  for Index := 1 to CornerCount do
  begin
    XString := ReadCharacters(24);
    YString := ReadCharacters(24);
    {$IFDEF WriteB} WriteLn(DemOutput,'corners (X,Y): ' + XString + ', ' + YString); {$ENDIF}
    X := FortranStrToFloat(XString);
    Y := FortranStrToFloat(YString);

    CornerPoint := TRbwZoomPoint.Create(RbwZoomBox1);
    try
      CornerPoint.X := X;
      CornerPoint.Y := Y;
      Corners.Add(CornerPoint);
    except
      CornerPoint.Free;
      raise;
    end;


  end;

  if CoordInSec and (Corners.Count > 0) then
  begin
    if GetCentralMeridian then
    begin
      X := 0;
      Y := 0;
      for Index := Corners.Count - CornerCount to Corners.Count -1 do
      begin
        CornerPoint := Corners[Index] as TRbwZoomPoint;
        X := X + CornerPoint.X;
        Y := Y + CornerPoint.Y;
      end;
      X := X/CornerCount;
      Y := Y/CornerCount;
      CentralMeridianRadians := UTMCentralMeridianRadians(Y,X);
    end;

  end;
  AString := ReadCharacters(24);
  {$IFDEF WriteB} WriteLn(DemOutput,'min or max elevation: ' + AString); {$ENDIF}
  MinOrMax := FortranStrToFloat(AString);
  RangeList.Add(MinOrMax);

  AString := ReadCharacters(24);
  {$IFDEF WriteB} WriteLn(DemOutput,'min or max elevation: ' + AString); {$ENDIF}
  MinOrMax := FortranStrToFloat(AString);
  RangeList.Add(MinOrMax);

  AString := ReadCharacters(24);
  Angle := FortranStrToFloat(AString);
  {$IFDEF WriteB} WriteLn(DemOutput,'angle (radians): ' + AString); {$ENDIF}
  {$IFDEF WriteB} WriteLn(DemOutput,'accuracy code: ' + {$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  AString := ReadCharacters(12);
  ResolutionX := FortranStrToFloat(AString);
  {$IFDEF WriteB} WriteLn(DemOutput,'resolution X: ' + AString); {$ENDIF}
  AString := ReadCharacters(12);
  ResolutionY := FortranStrToFloat(AString);
  {$IFDEF WriteB} WriteLn(DemOutput,'resolution Y: ' + AString); {$ENDIF}
  AString := ReadCharacters(12);
  ResolutionZ := FortranStrToFloat(AString);
  {$IFDEF WriteB} WriteLn(DemOutput,'resolution Z: ' +   AString  )  ; {$ENDIF}
  {$IFDEF WriteB} WriteLn(DemOutput,'rows: ' + {$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} ) {$ENDIF} ;
  ColumncountString := ReadCharacters(6);
  ColumnCount := StrToInt(ColumncountString);
  {$IFDEF WriteB} WriteLn(DemOutput,'columns: ' + ColumncountString); {$ENDIF}
  {$IFDEF WriteB} WriteLn(DemOutput,'Largest primary contour interval: ' + {$ENDIF}  ReadCharacters(5) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'source contour interval units: ' + {$ENDIF}  ReadCharacters(1) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Smallest primary contour interval: ' + {$ENDIF}  ReadCharacters(5) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'source contour interval units: ' + {$ENDIF}  ReadCharacters(1) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Data source date: ' + {$ENDIF}  ReadCharacters(4) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Data inspection/revision date: ' + {$ENDIF}  ReadCharacters(4) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Inspection/revision flag: ' + {$ENDIF}  ReadCharacters(1) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Data validation flag: ' + {$ENDIF}  ReadCharacters(1) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Suspect and void area flag: ' + {$ENDIF}  ReadCharacters(2) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Vertical datum: ' + {$ENDIF}  ReadCharacters(2) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Horizontal datum: ' + {$ENDIF}  ReadCharacters(2) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Data Edition: ' +  {$ENDIF} ReadCharacters(4) {$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'Percent Void: ' + {$ENDIF}  ReadCharacters(4) {$IFDEF WriteB} ) {$ENDIF} ;
  ReadRestOfRecord;
end;

function TfrmDEM2Image.FortranStrToFloat(AString : string) : double;
var
  DPos : integer;
  Sub : string;
begin
  AString := Trim(AString);
  DPos := Pos('d', AString);
  if DPos > 0 then
  begin
    AString[DPos] := 'e';
  end;
  DPos := Pos('D', AString);
  if DPos > 0 then
  begin
    AString[DPos] := 'E';
  end;
  if DecimalSeparator <> '.' then
  begin
    DPos := Pos(DecimalSeparator, AString);
    if DPos > 0 then
    begin
      AString[DPos] := DecimalSeparator;
    end;
  end;
  Sub := Copy(AString, 2, Length(AString));
  DPos := Pos('+', Sub);
  if DPos > 0 then
  begin
    if (AString[DPos] <> 'e') and (AString[DPos] <> 'E') then
    begin
      AString := Copy(AString, 1, DPos) + 'E' + Copy(AString, DPos + 1, Length(AString))
    end;
  end;
  DPos := Pos('-', Sub);
  if DPos > 0 then
  begin
    if (AString[DPos] <> 'e') and (AString[DPos] <> 'E') then
    begin
      AString := Copy(AString, 1, DPos) + 'E' + Copy(AString, DPos + 1, Length(AString))
    end;
  end;
  result := StrToFloat(AString);
end;

procedure TfrmDEM2Image.ReadRecordB;
var
  m, n : integer;
  mString, nString : string;
  Index : integer;
  FirstX, FirstY : double;
  FirstXString, FirstYString, DatumString : String;
  X, Y : double;
  AString : string;
  ElevationPoint : TElevationPoint;
  FirstPoint, SecondPoint : TElevationPoint;
  Datum : double;
begin
  FirstPoint := nil;
  CharCount := 0;
  {$IFDEF WriteB} WriteLn(DemOutput,''); {$ENDIF}
  {$IFDEF WriteB} WriteLn(DemOutput,'row: ' +{$ENDIF} ReadCharacters(6) {$IFDEF WriteB} ){$ENDIF};
  {$IFDEF WriteB} WriteLn(DemOutput,'column: ' +{$ENDIF} ReadCharacters(6) {$IFDEF WriteB} ){$ENDIF};
  mString := ReadCharacters(6);
  {$IFDEF WriteB} WriteLn(DemOutput,'m: ' + mString); {$ENDIF}
  nString := ReadCharacters(6);
  {$IFDEF WriteB} WriteLn(DemOutput,'n: ' + nString); {$ENDIF}
  FirstXString := ReadCharacters(24);
  {$IFDEF WriteB} WriteLn(DemOutput,'First X: ' + FirstXString); {$ENDIF}
  FirstYString := ReadCharacters(24);
  {$IFDEF WriteB} WriteLn(DemOutput,'First Y: ' + FirstYString); {$ENDIF}
  DatumString := ReadCharacters(24);
  {$IFDEF WriteB} WriteLn(DemOutput,'Datum: ' +  DatumString ) ; {$ENDIF}
  {$IFDEF WriteB} WriteLn(DemOutput,'min or max: ' + {$ENDIF} ReadCharacters(24){$IFDEF WriteB} ) {$ENDIF};
  {$IFDEF WriteB} WriteLn(DemOutput,'min or max: ' + {$ENDIF} ReadCharacters(24){$IFDEF WriteB} ) {$ENDIF};
  FirstX := FortranStrToFloat(FirstXString);
  FirstY := FortranStrToFloat(FirstYString);
  Datum := FortranStrToFloat(DatumString);
  m := StrToInt(Trim(mString));
  n := StrToInt(Trim(nString));
  for Index := 1 to m * n do
  begin
    if CharCount > 1024 -6 then
    begin
      ReadRestOfRecord;
    end;
    if angle = 0 then
    begin
      X := FirstX;
      Y := FirstY + (Index-1) * ResolutionY;
    end
    else
    begin
      X := FirstX + (Index-1) * ResolutionX * Sin(angle);
      Y := FirstY + (Index-1) * ResolutionY * Cos(angle);
    end;

    if CoordInSec then
    begin
      GetCoordinates(Y/60/60/180*PI, X/60/60/180*PI, CentralMeridianRadians, X, Y);
    end;

    AString := ReadCharacters(6);
    {$IFDEF WriteB} WriteLn(DemOutput,'X,Y,Elevation ' + IntToStr(Index) + ' : '
      + FloatToStr(X) + ', ' +  FloatToStr(Y) + ', ' +  AString); {$ENDIF}

    if Index =1 then
    begin
      FirstPoint := TElevationPoint.Create(RbwZoomBox1);
      try
        FirstPoint.UseForZoomOut := False;
        FirstPoint.Elevation := FortranStrToFloat(AString)*ResolutionZ+Datum;
        FirstPoint.X := X;
        FirstPoint.Y := Y;
      except
        FirstPoint.Free;
        raise;
      end;
    end
    else if Index = 2 then
    begin
      SecondPoint := TElevationPoint.Create(RbwZoomBox1);
      try
        SecondPoint.UseForZoomOut := False;
        SecondPoint.Elevation := FortranStrToFloat(AString)*ResolutionZ+Datum;
        SecondPoint.X := X;
        SecondPoint.Y := Y;
        PointSize := Abs(SecondPoint.YCoord - FirstPoint.YCoord) + 1;
        FirstPoint.Draw(DEM_Map, PointSize);
        SecondPoint.Draw(DEM_Map, PointSize);
      finally
        FirstPoint.Free;
        SecondPoint.Free;
      end;
    end
    else
    begin

      ElevationPoint := TElevationPoint.Create(RbwZoomBox1);
      try
        ElevationPoint.UseForZoomOut := False;
        ElevationPoint.Elevation := FortranStrToFloat(AString)*ResolutionZ+Datum;
        ElevationPoint.X := X;
        ElevationPoint.Y := Y;
        ElevationPoint.Draw(DEM_Map, PointSize);
      finally
        ElevationPoint.Free;
      end;
    end;

  end;
  ReadRestOfRecord

end;

procedure TfrmDEM2Image.ReadRecordC;
begin
  Try
  {$IFDEF WriteB} WriteLn(DemOutput,'Code indicating availability of statistics in data element 2: ' +{$ENDIF}  ReadCharacters(6){$IFDEF WriteB}) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'RMSE of file''s datum relative to absolute datum (x, y, z): '
    +{$ENDIF}  ReadCharacters(6) {$IFDEF WriteB} + ', ' + {$ELSE} ; {$ENDIF} ReadCharacters(6) {$IFDEF WriteB} + ', ' +{$ELSE} ; {$ENDIF} ReadCharacters(6) {$IFDEF WriteB}){$ENDIF};
  {$IFDEF WriteB} WriteLn(DemOutput,'sample size: ' + {$ENDIF} ReadCharacters(6){$IFDEF WriteB}){$ENDIF};
  {$IFDEF WriteB} WriteLn(DemOutput,'Code indicating availability of statistics in data element 5: ' +{$ENDIF} ReadCharacters(6){$IFDEF WriteB} ) {$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'RMSE of file''s datum relative to file''s datum (x, y, z): '
    +{$ENDIF} ReadCharacters(6) {$IFDEF WriteB} + ', ' +{$ELSE} ; {$ENDIF} ReadCharacters(6) {$IFDEF WriteB} + ', ' +{$ELSE} ; {$ENDIF} ReadCharacters(6){$IFDEF WriteB} ){$ENDIF} ;
  {$IFDEF WriteB} WriteLn(DemOutput,'sample size: ' + {$ENDIF} ReadCharacters(6) {$IFDEF WriteB} ){$ENDIF} ;
  except on E : EInOutError do
    begin
//      record C is not present.
    end;
  end;
end;

function TfrmDEM2Image.BM_Width : integer;
begin
  if not Done then
  begin
    result := RbwZoomBox1.PBWidth - RbwZoomBox1.LeftMargin
      - RbwZoomBox1.RightMargin;
  end
  else
  begin
    result := Abs(RbwZoomBox1.XCoord(RbwZoomBox1.MaxX)
      - RbwZoomBox1.XCoord(RbwZoomBox1.MinX));
  end;
end;

function TfrmDEM2Image.BM_Height : integer;
begin
  if not Done then
  begin
    result := RbwZoomBox1.PBHeight - RbwZoomBox1.TopMargin
      - RbwZoomBox1.BottomMargin;
  end
  else
  begin
    result := Abs(RbwZoomBox1.YCoord(RbwZoomBox1.MaxY)
      - RbwZoomBox1.YCoord(RbwZoomBox1.MinY))
  end;
end;

procedure TfrmDEM2Image.DrawOnBitMap;
var
  ARect : TRect;
begin
  DEM_Map.Height := BM_Height;
  DEM_Map.Width :=  BM_Width;
  ARect.Left := 0;
  ARect.Top := 0;
  ARect.Right := DEM_Map.Width;
  ARect.Bottom := DEM_Map.Height;
  DEM_Map.Canvas.Brush.Color := RbwZoomBox1.PBColor;
  DEM_Map.Canvas.FillRect(ARect);

  RbwZoomBox1.Invalidate;
end;

procedure TfrmDEM2Image.MakeBitMap(const ShouldZoomOut : boolean);
var
  Index : integer;
  FileIndex : integer;
  MeridianList : TRealList;
begin
  Open1.Enabled := False;
  Refresh1.Enabled := False;
  Save1.Enabled := False;
  sbZoom.Enabled := False;
  sbZoomIn.Enabled := False;
  sbZoomOut.Enabled := False;
  sbZoomExtents.Enabled := False;
  btnCancel.Enabled := True;
  cbWhiteOcean.Enabled := False;
  SaveasJPEG1.Enabled := False;
  cbRed.Enabled := False;
  CancelBitMap := False;
  RangeList.Clear;
  try

    DrawOnBitMap;

    Corners.Clear;
    RangeList.Clear;
    MeridianList := TRealList.Create;
    try
      for FileIndex := 0 to OpenDialog1.Files.Count -1 do
      begin
        StatusBar1.Panels[1].Text := 'Reading ' + ExtractFileName(OpenDialog1.Files[FileIndex]);
        AssignFile(DemFile, OpenDialog1.Files[FileIndex]);
        try
          Reset(DemFile, SizeOf(Char));
          ReadRecordA(True);
          if CoordInSec and (MeridianList.IndexOf(CentralMeridianRadians) < 0) then
          begin
            MeridianList.Add(CentralMeridianRadians);
          end;

        finally
          CloseFile(DemFile);
        end;
      end;

      if MeridianList.Count > 1 then
      begin
        for Index := 0 to MeridianList.Count -1 do
        begin
          frmCentralMeridian.rgCentralMeridians.Items.Add(
            FloatToStr(MeridianList[Index]*180/Pi));
        end;
        if not frmCentralMeridian.ShowModal = mrOK then
        begin
          Exit;
        end;
        CentralMeridianRadians := MeridianList[frmCentralMeridian.rgCentralMeridians.ItemIndex];

      end;
      GetRange;
      SetCornerCoordinates;
      if ShouldZoomOut then
      begin
        RbwZoomBox1.ZoomOut;
      end
      else
      begin
        RbwZoomBox1.GetMinMax;
      end;

      Done := True;
      DrawOnBitMap;

      Corners.Clear;

      for FileIndex := 0 to OpenDialog1.Files.Count -1 do
      begin
        StatusBar1.Panels[1].Text := 'Reading ' + ExtractFileName(OpenDialog1.Files[FileIndex]);
        AssignFile(DemFile, OpenDialog1.Files[FileIndex]);
    {$IFDEF WriteB}
        AssignFile(DemOutput, SaveDialog1.FileName);
    {$ENDIF}

        try
          Reset(DemFile, SizeOf(Char));
    {$IFDEF WriteB}
          Rewrite(DemOutput);
    {$ENDIF}
          ReadRecordA(False);
          for Index := 1 to ColumnCount do
          begin
            if CancelBitMap then Exit;
            ReadRecordB;
            Application.ProcessMessages;
            if Index mod 20 = 0 then
            begin
              RbwZoomBox1.Invalidate;
            end;

          end;
          ReadRecordC;

        finally
          CloseFile(DemFile);
    {$IFDEF WriteB}
          CloseFile(DemOutput);
    {$ENDIF}
        end;
      end;
      SetCornerCoordinates;
    finally
      MeridianList.Free;
    end;
  finally
    Open1.Enabled := True;
    Refresh1.Enabled := True;
    sbZoom.Enabled := True;
    sbZoomIn.Enabled := True;
    sbZoomOut.Enabled := True;
    sbZoomExtents.Enabled := True;
    btnCancel.Enabled := False;
    cbWhiteOcean.Enabled := True;
    Save1.Enabled := True;
    SaveasJPEG1.Enabled := True;
    cbRed.Enabled := True;
    StatusBar1.Panels[1].Text := 'Done';
  end;
end;

function TfrmDEM2Image.UTMCentralMeridianRadians(LatitudeSeconds,
  LongitudeSeconds: double): double;
var
  LongAngleDegrees : integer;
  CenterMeridianDegrees : integer;
begin
  LongAngleDegrees := Trunc(LongitudeSeconds/3600);
  CenterMeridianDegrees := LongAngleDegrees div 6;
  CenterMeridianDegrees := CenterMeridianDegrees * 6;
  if CenterMeridianDegrees < 0 then
  begin
    CenterMeridianDegrees := CenterMeridianDegrees - 3;
  end
  else
  begin
    CenterMeridianDegrees := CenterMeridianDegrees + 3;
  end;
  result := CenterMeridianDegrees*Pi/180;
end;

function TfrmDEM2Image.Get_N(LatitudeRadians: double): double;
begin
  result := LowerCase_a/Sqrt(1- e_square*Sqr(Sin(LatitudeRadians)));
end;

function TfrmDEM2Image.Get_T(LatitudeRadians: double): double;
begin
  result := Sqr(Tan(LatitudeRadians));
end;

function TfrmDEM2Image.Get_C(LatitudeRadians: double): double;
begin
  result := e_prime_square*Sqr(Cos(LatitudeRadians));
end;

function TfrmDEM2Image.Get_A(LongitudeRadians, LatitudeRadians,
  CentralMeridianRadians: double): double;
begin
  Result := (LongitudeRadians - CentralMeridianRadians) * Cos(LatitudeRadians);
end;

function TfrmDEM2Image.Get_M(LatitudeRadians: double): double;
var
  e_fourth, e_sixth : double;
begin
  e_fourth := Sqr(e_square);
  e_sixth := Power(e_square,3);
  Result := LowerCase_a *((1-e_square/4 - 3*e_fourth/64 - 5 * e_sixth/256)
    * LatitudeRadians - (3*e_square/8 + 3*e_fourth/32 + 45*e_sixth/1024)
    * sin(2*LatitudeRadians) + (15*e_fourth/256 + 45*e_sixth/1024)
    * sin(4*LatitudeRadians) - (35*e_sixth/3072) * sin(6*LatitudeRadians));
end;

procedure TfrmDEM2Image.GetCoordinates(const LatitudeRadians,
  LongitudeRadians, CentralMeridian: double; var X, Y: double);
const
  M_zero = 0;
var
  N, A, T, C, M  : double;
begin
  N := Get_N(LatitudeRadians);
  A := Get_A(LongitudeRadians, LatitudeRadians, CentralMeridian);
  T := Get_T(LatitudeRadians);
  C := Get_C(LatitudeRadians);
  M := Get_M(LatitudeRadians);

  X := k_zero * N *(A + (1 - T +C)*Power(A,3)/6 + (5 -18*T + Sqr(T)
    + 72*C - 58*e_prime_square)* Power(A,5)/120);

  Y := k_zero * (M - M_zero + N * Tan(LatitudeRadians) *
    (Sqr(A)/2 + (5 - T + 9*C + 4*Sqr(C)) * Power(A,4)/24 +
    (61 - 58* T + Sqr(T) + 600 * C - 330*e_prime_square)*Power(A,6)/720));

  X := X + 500000;

  if Y < 0 then
  begin
    Y := Y + 10000000
  end;
end;

procedure TfrmDEM2Image.FormCreate(Sender: TObject);
begin
  Constraints.MinWidth := Width;
  Corners := TObjectList.Create;
  DEM_Map := TBitMap.Create;
  DEM_Map.PixelFormat := pf24bit;
  RangeList := TRealList.Create;
  DrawOnBitMap;
  PointSize := 1;
end;

procedure TfrmDEM2Image.FormDestroy(Sender: TObject);
begin
  Corners.Free;
  DEM_Map.Free;
  RangeList.Free;
end;

procedure TfrmDEM2Image.RbwZoomBox1Paint(Sender: TObject);
var
  Rect : TRect;
  HeightRatio, WidthRatio : double;
begin
  if (Dem_Map.Height = BM_Height)
    and (Dem_Map.Width = BM_Width) then
  begin
    RbwZoomBox1.PBCanvas.Draw(RbwZoomBox1.LeftMargin,
      RbwZoomBox1.TopMargin, DEM_Map);
  end
  else
  begin
    Rect.Left:= RbwZoomBox1.LeftMargin;
    Rect.Top := RbwZoomBox1.TopMargin;

    HeightRatio := DEM_Map.Height/BM_Height;
    WidthRatio :=  DEM_Map.Width/BM_Width;
    if HeightRatio > WidthRatio then
    begin
      Rect.Bottom := Rect.Top + BM_Height;
      Rect.Right := Rect.Left + Round(Dem_Map.Width/HeightRatio);
    end
    else
    begin
      Rect.Right := Rect.Left + BM_Width;
      Rect.Bottom := Rect.Top + Round(Dem_Map.Height/WidthRatio);
    end;
    RbwZoomBox1.PBCanvas.StretchDraw(Rect,DEM_Map);
  end;  


end;

{ TElevationPoint }

procedure TElevationPoint.Draw(const ABitMap : TBitmap;
  const PointSize : integer);
var
  Rect: TRect;
  BM_X, BM_Y : integer;
  OffSet : integer;
begin
  OffSet := PointSize div 2;
  ABitMap.Canvas.Brush.Color := FColor;
  BM_X := XCoord - ZoomBox.LeftMargin - OffSet;
  BM_Y := YCoord - ZoomBox.TopMargin - OffSet;
  Rect.Top := BM_Y;
  Rect.Left := BM_X;
  Rect.Bottom := BM_Y+PointSize;
  Rect.Right := BM_X+ PointSize;
  ABitMap.Canvas.FillRect(Rect);
end;

procedure TElevationPoint.SetElevation(const Value: double);
var
  Fraction : double;
begin
  FElevation := Value;

  if frmDEM2Image.Range[1] = frmDEM2Image.Range[0] then
  begin
    FColor := 0;
  end
  else
  begin
    if frmDEM2Image.cbWhiteOcean.Checked and (Value = 0) then
    begin
      FColor := frmDEM2Image.RbwZoomBox1.PBColor;
    end
    else
    begin
      Fraction := (Elevation - frmDEM2Image.Range[0])
        /(frmDEM2Image.Range[1] - frmDEM2Image.Range[0]);
      if frmDEM2Image.cbRed.Checked then
      begin
        Fraction := 1 - Fraction;
      end;
      FColor := FracToRainbow(Fraction);
    end;
  end;

end;

procedure TfrmDEM2Image.sbZoomExtentsClick(Sender: TObject);
begin
  RbwZoomBox1.ZoomOut;
  sbPan.Enabled := False;
end;

procedure TfrmDEM2Image.sbZoomInClick(Sender: TObject);
begin
  RbwZoomBox1.ZoomBy(2);
  sbPan.Enabled := True;
end;

procedure TfrmDEM2Image.sbZoomOutClick(Sender: TObject);
begin
  RbwZoomBox1.ZoomBy(0.5);
end;

procedure TfrmDEM2Image.RbwZoomBox1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if sbZoom.Down then
  begin
    RbwZoomBox1.BeginZoom(X,Y);
  end
  else if sbPan.Down then
  begin
    RbwZoomBox1.BeginPan
  end;

end;

procedure TfrmDEM2Image.RbwZoomBox1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if sbZoom.Down then
  begin
    RbwZoomBox1.FinishZoom(X,Y);
    sbPan.Enabled := True;
  end
  else if sbPan.Down then
  begin
    RbwZoomBox1.EndPan;
  end;
end;

procedure TfrmDEM2Image.RbwZoomBox1MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  if sbZoom.Down then
  begin
    RbwZoomBox1.ContinueZoom(X,Y);
  end;
  StatusBar1.Panels[0].Text := '(X,Y) = ' + FloatToStr(RbwZoomBox1.X(X)) + ', '
    + FloatToStr(RbwZoomBox1.Y(Y));
end;

procedure TfrmDEM2Image.btnRefreshClick(Sender: TObject);
begin
  MakeBitmap(False);
end;

procedure TfrmDEM2Image.btnCancelClick(Sender: TObject);
begin
  CancelBitMap := True;
end;

procedure TfrmDEM2Image.Open1Click(Sender: TObject);
begin
  if OpenDialog1.Execute {$IFDEF WriteB} and SaveDialog1.Execute {$ENDIF} then
  begin
    MakeBitMap(True);
  end;
end;

procedure TfrmDEM2Image.Refresh1Click(Sender: TObject);
begin
  MakeBitmap(False);
end;

procedure TfrmDEM2Image.sbZoomClick(Sender: TObject);
begin
  if sbZoom.Down then
  begin
    RbwZoomBox1.Cursor := crCross;
    RbwZoomBox1.PBCursor := crCross;
    RbwZoomBox1.SCursor := crCross;
  end
  else
  begin
    RbwZoomBox1.Cursor := crDefault;
    RbwZoomBox1.PBCursor := crDefault;
    RbwZoomBox1.SCursor := crDefault;
  end;
end;

procedure TfrmDEM2Image.sbPanClick(Sender: TObject);
begin
  if sbPan.Down then
  begin
    RbwZoomBox1.Cursor := crHandPoint;
    RbwZoomBox1.PBCursor := crHandPoint;
    RbwZoomBox1.SCursor := crHandPoint;
  end
  else
  begin
    RbwZoomBox1.Cursor := crDefault;
    RbwZoomBox1.PBCursor := crDefault;
    RbwZoomBox1.SCursor := crDefault;
  end;

end;

procedure TfrmDEM2Image.Save1Click(Sender: TObject);
begin
  SaveDialog1.FileName := '';
  if SaveDialog1.Execute then
  begin
    DEM_Map.SaveToFile(SaveDialog1.FileName);
  end;
end;

procedure TfrmDEM2Image.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TfrmDEM2Image.About1Click(Sender: TObject);
begin
  frmAbout.ShowModal;
end;

procedure TfrmDEM2Image.Help2Click(Sender: TObject);
begin
  Application.HelpContext(10);
end;

procedure TfrmDEM2Image.SaveasJPEG1Click(Sender: TObject);

begin
  frmJpegChoices.SaveasJPEG(DEM_Map);
end;

procedure TfrmDEM2Image.FormResize(Sender: TObject);
begin
  StatusBar1.Panels[0].Width := StatusBar1.Width div 2;
end;

end.


